@@ -4,7 +4,6 @@ use warnings;
 use strict;
 
 use Test::Builder::Module;
-use Test::Aggregate::Builder;
 use Test::More;
 use Carp 'croak';
 
@@ -33,11 +32,11 @@ Test::Aggregate - Aggregate C<*.t> tests to make them run faster.
 
 =head1 VERSION
 
-Version 0.33
+Version 0.35
 
 =cut
 
-$VERSION = '0.33';
+$VERSION = '0.35';
 
 =head1 SYNOPSIS
 
@@ -48,6 +47,8 @@ $VERSION = '0.33';
     } );
     $tests->run;
 
+    ok $some_data, 'Test::Aggregate also re-exports Test::More functions';
+
 =head1 DESCRIPTION
 
 B<WARNING>:  this is ALPHA code.  The interface is not guaranteed to be
@@ -74,12 +75,26 @@ your regular test directory (C<t/> is the standard):
  });
  $tests->run;
 
+ ok $some_data, 'Test::Aggregate also re-exports Test::More functions';
+
 Take your simplest tests and move them, one by one, into the new test
 directory and keep running the C<Test::Aggregate> program.  You'll find some
 tests will not run in a shared environment like this.  You can either fix the
 tests or simply leave them in your regular test directory.  See how this
 distribution's tests are organized for an example.
 
+Note that C<Test::Aggregate> also exports all exported functions from
+C<Test::More>, allowing you to run other tests after the aggregated tests have
+run.
+
+ use Test::Aggregate;
+ my $other_test_dir = 'aggregate_tests';
+ my $tests = Test::Aggregate->new( {
+    dirs => $other_test_dir
+ });
+ $tests->run;
+ ok !(-f 't/data/tmp.txt'), '... and our temp file should be deleted';
+
 Some tests cannot run in an aggregate environment.  These may include
 test for this with the C<< $ENV{TEST_AGGREGATE} >> variable:
 
@@ -113,11 +128,20 @@ following keys:
 
 =over 4
 
-=item * C<dirs> (mandatory)
+=item * C<dirs> (either this or C<tests> is mandatory)
 
 The directories to look in for the aggregated tests.  This may be a scalar
 value of a single directory or an array refernce of multiple directories.
 
+=item * C<tests> (either this or C<dirs> is mandatory)
+
+Instead of providing directories for the aggregated tests, you may supply an
+array reference with a list of tests to aggregate.  If both are supplied,
+these tests will be appended to the list of tests found in C<dirs>.
+
+The C<matching> parameter does not apply to test files identified with this
+key.
+
 =item * C<verbose> (optional, but strongly recommended)
 
 If set with a true value, each test programs success or failure will be
@@ -176,11 +200,24 @@ each test:
 
 This is the default behavior.
 
-=item * C<dry>
+=item * C<findbin> (optional)
+
+If supplied with a true value, this will cause FindBin::again() to be called
+before each test file.
+
+This is turned off by default.
+
+Note that older versions of FindBin (pre 1.47) sometimes get confused about
+where the bin directory is when I set C<$0>.  I don't know why, but this is a
+rarely used option and only happens pre 5.8 perl, so I'm not too worried about
+it.  Just keep it in mind.
+
+=item * C<dry> (optional)
 
 Just print the tests which will be run and the order they will be run in
 (obviously the order will be random if C<shuffle> is true).
 
+
 =item * C<tidy>
 
 If supplied a true value, attempts to run C<Perl::Tidy> on the source code.
@@ -229,14 +266,25 @@ sub _code_attributes {
 sub new {
     my ( $class, $arg_for ) = @_;
 
-    unless ( exists $arg_for->{dirs} ) {
-        Test::More::BAIL_OUT("You must supply 'dirs'");
+    unless ( exists $arg_for->{dirs} || exists $arg_for->{tests} ) {
+        Test::More::BAIL_OUT("You must supply 'dirs' or 'tests'");
+    }
+    if ( exists $arg_for->{tests} && 'ARRAY' ne ref $arg_for->{tests} ) {
+        Test::More::BAIL_OUT(
+            "Argument for Test::Aggregate 'tests' key must be an array reference"
+        );
     }
         
     $arg_for->{test_nowarnings} = 1 unless exists $arg_for->{test_nowarnings};
     $arg_for->{set_filenames}   = 1 unless exists $arg_for->{set_filenames};
+    $arg_for->{findbin}         = 1 unless exists $arg_for->{findbin};
     my $dirs = delete $arg_for->{dirs};
-    $dirs = [$dirs] if 'ARRAY' ne ref $dirs;
+    if ( defined $dirs ) {
+        $dirs = [$dirs] if 'ARRAY' ne ref $dirs;
+    }
+    else {
+        $dirs = [];
+    }
 
     my $matching = qr//;
     if ( $arg_for->{matching} ) {
@@ -266,17 +314,20 @@ sub new {
     } => $class;
     $self->{$_} = delete $arg_for->{$_} foreach (
         qw/
-        dump
+        check_plan
         dry
+        dump
+        findbin
         set_filenames
         shuffle
-        verbose
-        tidy
-        check_plan
         test_nowarnings
+        tests
+        tidy
+        verbose
         /,
         $class->_code_attributes
     );
+    $self->{tests} ||= [];
 
     if ( my @keys = keys %$arg_for ) {
         local $" = ', ';
@@ -304,17 +355,19 @@ we cannot load Data::Dump::Streamer:  $error.
 
 sub _check_plan      { shift->{check_plan} || 0 }
 sub _dump            { shift->{dump} || '' }
+sub _dry             { shift->{dry} }
 sub _should_shuffle  { shift->{shuffle} }
 sub _matching        { shift->{matching} }
 sub _set_filenames   { shift->{set_filenames} }
+sub _findbin         { shift->{findbin} }
 sub _dirs            { @{ shift->{dirs} } }
 sub _startup         { shift->{startup} }
 sub _shutdown        { shift->{shutdown} }
 sub _setup           { shift->{setup} }
 sub _teardown        { shift->{teardown} }
+sub _tests           { @{ shift->{tests} } }
 sub _tidy            { shift->{tidy} }
 sub _test_nowarnings { shift->{test_nowarnings} }
-sub _dry             { shift->{dry} }
 
 sub _verbose        {
     my $self = shift;
@@ -329,12 +382,15 @@ sub _get_tests {
     my $self = shift;
     my @tests;
     my $matching = $self->_matching;
-    find( {
-            no_chdir => 1,
-            wanted   => sub {
-                push @tests => $File::Find::name if /\.t\z/ && /$matching/;
-            }
-    }, $self->_dirs );
+    if ( $self->_dirs ) {
+        find( {
+                no_chdir => 1,
+                wanted   => sub {
+                    push @tests => $File::Find::name if /\.t\z/ && /$matching/;
+                }
+        }, $self->_dirs );
+    }
+    push @tests => $self->_tests;
     
     if ( $self->_should_shuffle ) {
         $self->_shuffle(@tests);
@@ -387,19 +443,25 @@ sub run {
     # this eval'ed code has a check to not actually run the tests if we are
     # not in the dump file.
     eval $code;
+
     if ( my $error = $@ ) {
         croak("Could not run tests: $@");
     }
 
     $self->_startup->() if $self->_startup;
     my $builder = Test::Builder->new;
-    
-    foreach my $data ($self->_packages) {
+
+    # some tests may have been run in BEGIN blocks
+    $builder->{TEST_AGGREGATE_last_test} = @{ $builder->{Test_Results} } || 0;
+
+    my $current_test = 0;
+    my @packages     = $self->_packages;
+    my $total_tests  = @packages;
+    foreach my $data (@packages) {
+        $current_test++;
         my ( $test, $package ) = @$data;
-        Test::More::diag("******** running tests for $test ********")
-          if $ENV{TEST_VERBOSE};
         $self->_setup->() if $self->_setup;
-        eval { $package->run_the_tests };
+        _run_this_test_program( $package => $test, $current_test, $total_tests, 2 );
         if ( my $error = $@ ) {
             Test::More::ok( 0, "Error running ($test):  $error" );
         }
@@ -413,6 +475,68 @@ sub run {
     $self->_shutdown->() if $self->_shutdown;
 }
 
+sub _any_tests_failed {
+    my $failed  = 0; 
+    my $builder = Test::Builder->new;
+    my @summary = $builder->summary;
+    foreach my $passed (@summary[ $builder->{TEST_AGGREGATE_last_test} ..
+        $builder->current_test - 1 ]) {
+        if (not $passed) {
+            $failed = 1;
+            last;
+        }
+    }
+    return $failed;
+}
+
+sub _run_this_test_program {
+    my $builder = Test::Builder->new;
+    my ( $package, $test, $current_test, $num_tests, $verbose ) = @_;
+    Test::More::diag("******** running tests for $test ********") if $ENV{TEST_VERBOSE};
+    my $error = eval { 
+        if ( my $reason = $Test::Aggregate::Builder::SKIP_REASON_FOR{$package} ) {
+            Test::Builder->new->skip($reason);
+            return;
+        }
+        else {
+            local $@;
+            # localize some popular globals
+            no warnings 'uninitialized';
+            local %ENV = %ENV;
+            local $/   = $/;
+            local @INC = @INC;
+            local $_   = $_;
+            local $|   = $|;
+            local %SIG = %SIG;
+            use warnings 'uninitialized';
+            $Test::Aggregate::Builder::FILE_FOR{$package} = $test;
+            eval { $package->run_the_tests };
+            $@;
+        }
+    };
+
+    if ($verbose) {
+        my $test_name = "$test ($current_test out of $num_tests)";
+        my $failed    = _any_tests_failed();
+        chomp $error if defined $error;
+        $error &&= "($error)";
+        my $ok = $failed || $error
+                ? "not ok - $test_name $error"
+                : "    ok - $test_name";
+        Test::More::diag($ok) if $error or $failed or $verbose == $VERBOSE{all};
+        if ($error or $failed) {
+            Test::More::ok(0, "Error running ($test):  $error");
+            # XXX this should be fine since these keys are not actually used
+            # internally.
+            $builder->{XXX_test_failed}       = 0;
+            $builder->{TEST_MOST_test_failed} = 0;
+        }
+    }
+    $builder->{TEST_AGGREGATE_last_test} = $builder->current_test;
+
+    return unless $error;
+}
+
 sub _build_aggregate_code {
     my ( $self, @tests ) = @_;
     my $code = $self->_test_builder_override;
@@ -421,14 +545,24 @@ sub _build_aggregate_code {
     my ( $shutdown, $shutdown_code ) = $self->_as_code('shutdown');
     my ( $setup,    $setup_code )    = $self->_as_code('setup');
     my ( $teardown, $teardown_code ) = $self->_as_code('teardown');
+    
     my $verbose = $self->_verbose;
-
+    my $findbin;
+    if ( $self->_findbin ) {
+        $findbin = <<'        END_CODE';
+use FindBin;
+my $REINIT_FINDBIN = FindBin->can(q/again/) || sub {};
+        END_CODE
+    }
+    else {
+        $findbin = 'my $REINIT_FINDBIN = sub {};';
+    }
     $code .= <<"    END_CODE";
 $startup_code
 $shutdown_code
 $setup_code
 $teardown_code
-my \$LAST_TEST_NUM = 0;
+$findbin
     END_CODE
     
     my @packages;
@@ -438,11 +572,7 @@ my \$LAST_TEST_NUM = 0;
 
     my $dump = $self->_dump;
 
-    $code .= <<"    END_CODE";
-if ( __FILE__ eq '$dump' ) {
-    package Test::Aggregate; # ;)
-    my \$builder = Test::Builder->new;
-    END_CODE
+    $code .= "if ( __FILE__ eq '$dump' ) {\n";
 
     if ( $startup ) {
         $code .= "    $startup->() if __FILE__ eq '$dump';\n";
@@ -465,26 +595,13 @@ if ( __FILE__ eq '$dump' ) {
         if ( $test_code =~ /^(__(?:DATA|END)__)/m ) {
             Test::More::BAIL_OUT("Test $test not allowed to have $1 token");
         }
-        if ( $test_code =~ /skip_all/m ) {
-            warn
-              "Found possible 'skip_all'.  This can cause test suites to abort";
-        }
         my $package   = $self->_get_package($test);
         push @{ $self->{_packages} } => [ $test, $package ];
         if ( $setup ) {
             $code .= "    $setup->('$test');\n";
         }
-        $code .= <<"        END_CODE";
-    Test::More::diag("******** running tests for $test ********") if \$ENV{TEST_VERBOSE};
-    eval { $package->run_the_tests };
-    if ( my \$error = \$@ ) {
-        Test::More::ok( 0, "Error running ($test):  \$error" );
-        # XXX this should be fine since these keys are not actually used
-        # internally.
-        \$builder->{XXX_test_failed}       = 0;
-        \$builder->{TEST_MOST_test_failed} = 0;
-    }
-        END_CODE
+        $code .= qq{    Test::Aggregate::_run_this_test_program( $package => "$test", $current_test, $total_tests, $verbose );};
+
         if ( $teardown ) {
             $code .= "    $teardown->('$test');\n";
         }
@@ -493,36 +610,16 @@ if ( __FILE__ eq '$dump' ) {
         my $set_filenames = $self->_set_filenames
             ? "local \$0 = '$test';"
             : '';
-        my $see_if_tests_passed = $verbose ? <<"        END_CODE" : '';
-{
-    my \$builder = Test::Builder->new;   # singleton
-    my \$tests   = \$builder->current_test;
-    my \$failed = 0;
-    my \@summary = \$builder->summary;
-    foreach my \$passed ( \@summary[\$LAST_TEST_NUM .. \$tests - 1] ) {
-        if ( not \$passed ) {
-            \$failed = 1;
-            last;
-        }
-    }
-    my \$ok = \$failed ? "not ok - $test" : "    ok - $test ($current_test out of $total_tests)";
-    if ( \$failed or $verbose == $VERBOSE{all} ) {
-        Test::More::diag(\$ok);
-    }
-    \$LAST_TEST_NUM = \$tests;
-}
-        END_CODE
         $test_packages .= <<"        END_CODE";
 {
 $separator beginning of $test $separator
-package $package;
-sub run_the_tests {
-\$Test::Aggregate::Builder::FILE_FOR{$package} = '$test';
-$set_filenames
+    package $package;
+    sub run_the_tests {
+        $set_filenames
+        \$REINIT_FINDBIN->();
 # line 1 "$test"
 $test_code
-$see_if_tests_passed
-}
+    }
 $separator end of $test $separator
 }
         END_CODE
@@ -608,9 +705,12 @@ BEGIN {
     }
 
     return <<"    END_CODE";
+use Test::Aggregate;
 use Test::Aggregate::Builder;
 BEGIN { \$Test::Aggregate::Builder::CHECK_PLAN = $check_plan };
 $disable_test_nowarnings;
+
+my \$BUILDER = Test::Builder->new;
     END_CODE
 }
 
@@ -692,6 +792,35 @@ The teardown function gets run after every test program.
 
 =back
 
+=head1 GLOBAL VARIABLES
+
+You shouldn't be using global variables and a dependence on them can break
+your code.  However, Perl provides quite a few handy global variables which,
+unfortunately, can easily break your tests if you change them in one test and
+another assumes an unchanged value.  As a result, we localize many of Perl's
+most common global variables for you, using the following syntax:
+
+    local %ENV = %ENV; 
+    
+The following global variables are localized for you.  Any others must be
+localized manually per test.
+
+=over 4
+
+=item * C<@INC>
+
+=item * C<%ENV>
+
+=item * C<%SIG>
+
+=item * C<$/>
+
+=item * C<$_>
+
+=item * C<$|>
+
+=back
+
 =head1 CAVEATS
 
 Not all tests can be included with this technique.  If you have C<Test::Class>
@@ -745,12 +874,6 @@ This is needed when you use C<check_plan> and have C<Test::NoWarnings> used.
 This is because we do work internally to subtract the extra test added by
 C<Test::NoWarnings>.  It's painful and experimental.  Good luck.
     
-=item * No 'skip_all' tests, please
-
-Tests which potentially 'skip_all' will cause the aggregate test suite to
-abort prematurely.  Do not attempt to aggregate them.  This may be fixed in a
-future release.
-
 =item * C<Variable "$x" will not stay shared at (eval ...>
 
 Because each test is wrapped in a method call, any of your subs which access a
